<?php

/**
 * Info Indexmenu: Displays the index of a specified namespace. 
 *
 * @license     GPL 2 (http://www.gnu.org/licenses/gpl.html)
 * @author      Samuele Tognini <samuele@cli.di.unipi.it>
 * 
 */
 
if(!defined('DOKU_INC')) define('DOKU_INC',realpath(dirname(__FILE__).'/../../').'/');
if(!defined('DOKU_PLUGIN')) define('DOKU_PLUGIN',DOKU_INC.'lib/plugins/');
if(!defined('INDEXMENU_IMG_RELDIR')) define('INDEXMENU_IMG_RELDIR',DOKU_BASE.'lib/plugins/indexmenu/images');
if(!defined('INDEXMENU_IMG_ABSDIR')) define('INDEXMENU_IMG_ABSDIR',realpath(dirname(__FILE__)."/images"));
require_once(DOKU_PLUGIN.'syntax.php');
require_once(DOKU_INC.'inc/search.php');
 
/**
 * All DokuWiki plugins to extend the parser/rendering mechanism
 * need to inherit from this class
 */
class syntax_plugin_indexmenu extends DokuWiki_Syntax_Plugin {
  
  /**
   * return some info
   */
  function getInfo(){
    return array(
		 'author' => 'Samuele Tognini',
		 'email'  => 'samuele@cli.di.unipi.it',
		 'date'   => rtrim(io_readFile(DOKU_PLUGIN.'indexmenu/VERSION.txt')),
		 'name'   => 'Indexmenu',
		 'desc'   => 'Insert the index of a specified namespace.',
		 'url'    => 'http://wiki.splitbrain.org/plugin:indexmenu'
		 );
  }
  
  /**
   * What kind of syntax are we?
   */
  function getType(){
    return 'substition';
  }

  function getPType(){
    return 'block';
  }
 
  /**
   * Where to sort in?
   */
  function getSort(){
    return 138;
  }
 
  /**
   * Connect pattern to lexer
   */
  function connectTo($mode) {
    $this->Lexer->addSpecialPattern('{{indexmenu>.+?}}',$mode,'plugin_indexmenu');
  }
      
  /**
   * Handle the match
   */
  function handle($match, $state, $pos, &$handler){
    $theme="default/";
    $ns=".";
    $level = -1;
    $nons = true;
    $gen_id='random';
    $max=0;
    $nss=array();
    $sort='nosort';
    $match = substr($match,12,-2);
    //split namespace,level,theme
    $match = preg_split('/\|/u', $match, 2);
    //split optional namespaces
    $nss_temp=preg_split("/ /u",$match[0],-1,PREG_SPLIT_NO_EMPTY);
    //Array optional namespace => level
    for ($i = 1; $i < count($nss_temp); $i++) {
      $nsss=preg_split("/#/u",$nss_temp[$i]);
      $nss[]=array($this->_parse_ns($nsss[0]),(is_numeric($nsss[1])) ? $nsss[1] : $level);
    }
    //split main requested namespace
    if (preg_match('/(.*)#(\S*)/u',$nss_temp[0],$ns_opt)) {
      //split level
      $ns=$ns_opt[1];
      if (is_numeric($ns_opt[2])) $level=$ns_opt[2];
    } else {
      $ns=$nss_temp[0];
    }
    $ns=$this->_parse_ns($ns);
    $opts=preg_split('/ /u',$match[1]);
    //noscroll option
    $noscroll=in_array('noscroll',$opts);
    //Open at current namespace option
    $navbar=in_array('navbar',$opts);
    if ($navbar) $gen_id="1";
    //nocookie option
    $nocookie=in_array('nocookie',$opts);
    //plugin options
    $nons = in_array('nons',$opts);
    //disable toc preview
    $notoc = in_array('notoc',$opts);
    //sort option
    if (in_array('tsort',$opts)) $sort='tsort';
    //javascript option
    $js = in_array('js',$opts);
    if (!$js) {
      //split theme
      if (preg_match('/js#(\S*)/u',$match[1],$tmp_theme) > 0) {
	if (is_dir(INDEXMENU_IMG_ABSDIR."/".$tmp_theme[1])) {
	  $theme=$tmp_theme[1]."/";
	}
	$js=true;
      } 
    }
    $theme=INDEXMENU_IMG_RELDIR."/".$theme;
    //id generation method 
    if (preg_match('/id#(\S+)/u',$match[1],$id) >0) $gen_id=$id[1];
    //max option
    if (preg_match('/max#(\d+)/u',$match[1],$maxtmp) >0) $max=$maxtmp[1];
    //js options
    $js_opts=compact('theme','gen_id','nocookie','navbar','noscroll','max','notoc');
    return array($ns,
		 $js,
		 $js_opts,
		 $sort,
		 array('level' => $level,
		       'nons' => $nons,
		       'nss' => $nss,
		       'skip_index' => $this->getConf('skip_index'),
		       'skip_file' => $this->getConf('skip_file'),
		       'headpage' => $this->getConf('headpage'),
		       'hide_headpage' => $this->getConf('hide_headpage')
		       )
		 );
  }  
  
  /**
   * Render output
   */
  function render($mode, &$renderer, $data) {
    global $ACT;
    global $conf;
    global $INFO;
    if($mode == 'xhtml'){ 
      //Check user permission to display indexmenu in a preview page
      if ($ACT == 'preview' && 
	  $this->getConf('only_admins') && 
	  $conf['useacl'] &&
	  $INFO['perm'] < AUTH_ADMIN)
	return false;
      $n=$this->_indexmenu($data);
      if (!@$n) {
	$n = $this->getConf('empty_msg');
	$n= str_replace('{{ns}}',cleanID($data[0]),$n);
	$n=p_render('xhtml',p_get_instructions($n),$info);
      }
      $renderer->doc .= $n;
      return true;
    } else if ($mode == 'metadata') {
      //this is an indexmenu page;
      $renderer->meta['indexmenu'] = true;
      unset($renderer->persistent['indexmenu']);
      return true;
    } else {
      return false;
    }
  }
  
  /**
   * Return the index 
   * @author Samuele Tognini <samuele@cli.di.unipi.it>
   *
   * This function is a simple hack of Dokuwiki html_index($ns)
   * @author Andreas Gohr <andi@splitbrain.org>
   */
  function _indexmenu($myns) {
    global $conf;

    $ns = $myns[0];
    $js = $myns[1];
    $js_opts=$myns[2];
    $sort=$myns[3];
    $opts = $myns[4];
    $opts['js']=$js;
    $output=false;
    $data = array();
    $js_name=false;
    $fsdir="/".utf8_encodeFN(str_replace(':','/',$ns));
    if ($sort=='nosort') {
      search($data,$conf['datadir'],array($this,'_search_index'),$opts,$fsdir);
    } else {
      $this->_search($data,$conf['datadir'],array($this,'_search_index'),$opts,$fsdir);
    }
    if (!$data) return false;

    //javascript index
    if ($js) {      
      $ns = str_replace('/',':',$ns);
      $output_tmp=$this->_jstree($data,$ns,$js_opts,$js_name);
      //remove unwanted nodes from standard index 
      $this->_clean_data($data);
    }
    //Nojs dokuwiki index
    $output.="\n".'<div';
    if ($js_name) $output.=' id="nojs_'.$js_name.'" style="display:block;"';
    $output.=">\n";
    $output.=html_buildlist($data,'idx',array($this,"_html_list_index"),"html_li_index");
    $output.="</div>\n";
    $output.=$output_tmp;
    return $output;
  }

  /**
   * Build the browsable index of pages using javascript
   *
   * @author  Samuele Tognini <samuele@cli.di.unipi.it>
   */
  function _jstree($data,$ns,$js_opts,&$js_name) {
    global $conf;
    //Render requested ns as root
    $headpage=$this->getConf('headpage');
    if (empty($ns) && !empty($headpage)) $headpage .= ",".$conf['start'];
    $title=$this->_getTitle($ns,$headpage,$link);
    if (empty($ns)) (empty($title) && $conf['title']) ? $title=$conf['title'] : $title=$ns;
    $extra=array();
    if (empty($data)) return false;
    // Id generation method
    if (is_numeric($js_opts['gen_id'])) {
      $jsns=$js_opts['gen_id'];
    } elseif ($js_opts['gen_id'] == 'ns') {
      $jsns = sprintf("%u",crc32($ns));
    } else {
      $jsns=uniqid(rand());
    }
    $js_name="indexmenu_".$jsns;
    $out = "<script type='text/javascript' charset='utf-8'>\n";
    $out .= "var $js_name = new dTree('".$js_name."','".$js_opts['theme']."');\n";
    if ($js_opts['notoc']) $out .="$js_name.config.toc=false;\n";
    if ($js_opts['nocookie']) $out .="$js_name.config.useCookies=false;\n";
    if ($js_opts['noscroll']) $out .="$js_name.config.scroll=false;\n";
    if ($js_opts['max'] > 0)  $out .= "$js_name.config.max=".$js_opts['max'].";\n";
    $out .= $js_name.".add('".ltrim($ns,":")."',0,-1,'".$title."','".$link."');\n";
    //Array of nodes to check
    $q=array('0');
    //Current open node
    $node=0;
    foreach ($data as $i=>$item){
      $i++;
      $navbar="false";
      //Title and headpage configuration
      if ($item['type']=='d') {
	//Add information to open level at current namespace
	$navbar=($js_opts['navbar']) ? "'".$item['id']."'" : $navbar;
      }
      //Remove already processed nodes (greater level = lower level)
      while ($item['level'] <= $data[end($q)-1]['level']) {
	array_pop($q);  
      }

      //till i found its father node
      if ($item['level']==1) {
	//root node
	$father='0';
      } else {
	//Father node
	$father=end($q);
      }
      //add node and its options
      if ($item['type'] == 'd' ) {
	//find lowest open node of a tree branch in order to open it.
	if ($item['open']) ($item['level'] < $data[$node]['level']) ? $node=$i : $extra[]= $i;
	//insert node in last position
	array_push($q,$i);
      }
      $out .= $js_name.".add('".$item['id']."',$i,".$father.",'".$item['title']."'";
      if (!empty($item['link'])) $out .= ",'".$item['link']."'";
      $out .= ");\n";
    }
    $out .= "document.write(".$js_name.");\n";
    //Not all closed
    if ($node>0 || !empty($extra)) {
      //js open nodes
      $extra="addInitEvent(function(){".$js_name.".getOpenTo('".implode(",",$extra)."');});\n";
      $out .= $extra;
    }
    if ($js_opts['navbar']) $out .= "addInitEvent(function(){".$js_name.".openCurNS();});\n";
    if (!$js_opts['nocookie']) $out .= "addInitEvent(function(){".$js_name.".openCookies();});\n";
    $out .= "</script>\n";
    return $out;
  }

  /**
   * Get page title, checking for headpages
   *
   * @author  Samuele Tognini <samuele@cli.di.unipi.it>
   */
  function _getTitle (&$ns,$headpage,&$link) {
    global $conf;
    $link='';
    $title=noNS($ns);
    if (empty($headpage)) return $title;
    $ahp=explode(",",$headpage);
    foreach ($ahp as $hp) {
      switch ($hp) {
      case ":inside:":
	$page=$ns.":".noNS($ns);
	break;
      case ":same:":
	$page=$ns;
	break;
	//it's an inside start
      case ":start:":
	$page=$ns.":".$conf['start'];
	break;
	//inside pages
      default:
	$page=$ns.":".cleanID($hp);
      }
      resolve_pageid($ns, $page, $exists);
      //check headpage
      if (@file_exists(wikiFN($page)) && auth_quickaclcheck($page) >= AUTH_READ) {
	if ($conf['useheading'] && $title_tmp=p_get_metadata($page,'title')) $title=$title_tmp;
	$title=htmlspecialchars($title,ENT_QUOTES);
	$link=wl($page);
	$ns=$page;
	//headpage found, exit for
	break;
      }
    }
    return $title;
  }

  /**
   * Parse namespace request
   *
   * @author  Samuele Tognini <samuele@cli.di.unipi.it>
   */
  function _parse_ns ($ns) {
    global $ID;
    $ns=preg_replace("/^\.(:|$)/",dirname(str_replace(':','/',$ID))."$1",$ns);
    $ns=str_replace("/",":",$ns);
    $ns = cleanID($ns);
    return $ns;
  }
  
  /**
   * Clean index data from unwanted nodes in nojs mode.
   *
   * @author  Samuele Tognini <samuele@cli.di.unipi.it>
   */
  function _clean_data(&$data) {
    foreach ($data as $i=>$item) {
      //closed node
      if ($item['type'] == "d" && !$item['open']) {
	$a=$i+1;
	$level=$data[$i]['level'];
	//search and remove every lower and closed nodes
	while ($data[$a]['level'] > $level && !$data[$a]['open']) {
	  unset($data[$a]);
	  $a++;
	}
      }
      $i++;
    }
  }

  /**
   * Build the browsable index of pages
   *
   * $opts['ns'] is the current namespace
   *
   * @author  Andreas Gohr <andi@splitbrain.org>
   * modified by Samuele Tognini <samuele@cli.di.unipi.it>
   */
  function _search_index(&$data,$base,$file,$type,$lvl,$opts){
    global $conf;
    $isopen = false;
    $item = array();
    $skip_index=$opts['skip_index'];
    $skip_file=$opts['skip_file'];
    $headpage=$opts['headpage'];
    $id = pathID($file);
    if($type == 'd'){
      // Skip folders in plugin conf
      if (!empty($skip_index) &&
	  preg_match($skip_index, $file))
	return false;
      //open requested level
      if ($opts['level'] > $lvl || $opts['level'] == -1 ) $isopen=true;
      //check ACL (for namespaces too)
      if (auth_quickaclcheck($id.':*') < AUTH_READ) return false;
      //Not added, but it lets the check go on.
      if ($opts['nons']) return true;
      //Set title and link
      $title=$this->_getTitle($id,$headpage,$link);
    } elseif ($type == 'f') {
      //Nons.Set all pages at first level
      if ($opts['nons']) $lvl=1;    
      //don't add
      if (!preg_match('#\.txt$#',$file)) return false;
      //check hiddens and acl
      if (isHiddenPage($id) || auth_quickaclcheck($id) < AUTH_READ) return false;
      //Skip files in plugin conf
      if (!empty($skip_file) &&
	  preg_match($skip_file, $file))
	return false;
      //Skip headpages to hide
      if (!$opts['nons'] && 
	  !empty($headpage) && 
	  $opts['hide_headpage']) {
	if ($id==$conf['start']) return false;
	$ahp=explode(",",$headpage);
	foreach ($ahp as $hp) {
	  switch ($hp) {
	  case ":inside:":
	    if (noNS($id)==noNS(getNS($id)))  return false;
	    break;
	  case ":same:":
	    if (@is_dir(dirname(wikiFN($id))."/".utf8_encodeFN(noNS($id)))) return false;
	    break;
	    //it' s an inside start
	  case ":start:":
	    if (noNS($id)==$conf['start']) return false;
	    break;
	  default:
	    if (noNS($id)==cleanID($hp)) return false;
	  }
	}
      }
      //Set title and link
      if (!$conf['useheading'] || !$title=p_get_metadata($id,'title')) $title=noNS($id);
      $title=htmlspecialchars($title,ENT_QUOTES);
      $link=wl($id);
    }

    //Search optional namespaces
    if ($type == 'd' && !empty($opts['nss'])){
      $nss=$opts['nss'];
      for ($a=0; $a<count($nss);$a++) {
	if (preg_match("/^".$id."($|:.+)/i",$nss[$a][0],$match) || //It contains an optional namespace
	    preg_match("/^".$nss[$a][0]."(:.*)/i",$id,$match))     //it's inside an optional namespace
	  //Check level counting the ":" occurrences
	  ($nss[$a][1] == -1 || substr_count($match[1],":") < $nss[$a][1]) ? $isopen=true : $isopen=false;
      }
    }

    $data[]=array( 'id'    => $id,
		   'type'  => $type,
		   'level' => $lvl,
		   'open'  => $isopen, 
		   'title' => $title,
		   'link'  => $link
		   );

    //js needs all nodes
    if ($opts['js']) $isopen= true;
    return $isopen;
  }  


  /**
   * Index item formatter
   *
   * User function for html_buildlist()
   *
   * @author Andreas Gohr <andi@splitbrain.org>
   * modified by Samuele Tognini <samuele@cli.di.unipi.it>
   */
  function _html_list_index($item){
    $ret = '';
    $link= $item['link'];
    $title = $item['title'];
    $nspage=$item['id'];
    $headpage=$this->getConf('headpage');
    //namespace
    if($item['type']=='d'){
      //namespace link
      if (empty($link)) {
	//standard namespace
	$link=wl($item['title'],'idx='.$item['id']);
	if (!empty($headpage)) $tagid="indexmenu_idx";
      } else {
	//namespace with headpage
	$tagid="indexmenu_idx_head";
      }
      $ret .= '<a href="'.$link.'" class="'.$tagid.'">';
      $ret .= $title;
      $ret .= '</a>';
    }else{
      //page link
      $ret .= html_wikilink(':'.$item['id']);
    }  
    return $ret;
  }


  /**
   * recurse direcory
   *
   * This function recurses into a given base directory
   * and calls the supplied function for each file and directory
   *
   * @param   array ref $data The results of the search are stored here
   * @param   string    $base Where to start the search
   * @param   callback  $func Callback (function name or arayy with object,method)
   * @param   string    $dir  Current directory beyond $base
   * @param   int       $lvl  Recursion Level
   * @author  Andreas Gohr <andi@splitbrain.org>
   * modified by Samuele Tognini <samuele@cli.di.unipi.it>
   */
  function _search(&$data,$base,$func,$opts,$dir='',$lvl=1){
    $dirs   = array();
    $files  = array();
    $data_tmp=array();

    //read in directories and files
    $dh = @opendir($base.'/'.$dir);
    if(!$dh) return;
    while(($file = readdir($dh)) !== false){
      if(preg_match('/^[\._]/',$file)) continue; //skip hidden files and upper dirs
      if(is_dir($base.'/'.$dir.'/'.$file)){
	$dirs[] = $dir.'/'.$file;
	continue;
      }elseif(substr($file,-5) == '.lock'){
	//skip lockfiles
	continue;
      }
      $files[] = $dir.'/'.$file;
    }
    closedir($dh);
    sort($dirs);
    //give directories to userfunction then recurse
    foreach($dirs as $dir){
      if (search_callback($func,$data,$base,$dir,'d',$lvl,$opts)){
	$this->_search($data,$base,$func,$opts,$dir,$lvl+1);
      }
    }
    //now handle the files
    foreach($files as $file){
      search_callback($func,$data_tmp,$base,$file,'f',$lvl,$opts);
    }
    usort($data_tmp,array($this,"_cmp"));
    $data=array_merge($data,$data_tmp);
  }

  /**
   * Sort nodes
   *
   */
  function _cmp($a, $b) {
    return strnatcasecmp($a['title'], $b['title']);
  }
} //Indexmenu class end  
